/**
 * @license
 * Copyright 2019 Google LLC. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * =============================================================================
 */
/// <reference types="@webgpu/types" />
import './flags_webgpu';
import { backend_util, DataStorage, DataType, KernelBackend, RecursiveArray, TensorInfo, TimingInfo } from '@tensorflow/tfjs-core';
import { Glslang } from '@webgpu/glslang/dist/web-devel/glslang.onefile';
import { BufferManager } from './buffer_manager';
import { FromPixelsProgram } from './kernels/FromPixels_utils/from_pixels_webgpu';
import * as webgpu_program from './kernels/webgpu_program';
export interface WebGPUMemoryInfo extends backend_util.MemoryInfo {
    numBytesInGPU: number;
    numBytesAllocatedInGPU: number;
    unreliable: boolean;
}
declare type BufferInfo = {
    byteSize: number;
    usage: GPUBufferUsageFlags;
    buffer?: GPUBuffer;
};
declare type TensorBufferInfo = {
    values: backend_util.BackendValues;
    dtype: DataType;
    bufferInfo: BufferInfo;
    refCount: number;
    complexTensorInfos?: {
        real: TensorInfo;
        imag: TensorInfo;
    };
};
interface DataId {
}
export declare type WebGPUKernelInfo = {
    name: string;
    query: Promise<number>;
};
export declare type TimerNode = RecursiveArray<WebGPUKernelInfo> | WebGPUKernelInfo;
export interface WebGPUTimingInfo extends TimingInfo {
    uploadWaitMs: number;
    downloadWaitMs: number;
}
export declare class WebGPUBackend extends KernelBackend {
    device: GPUDevice;
    queue: GPUQueue;
    glslang: Glslang;
    currentCommandEncoder: GPUCommandEncoder;
    tensorMap: DataStorage<TensorBufferInfo>;
    fromPixelProgram: FromPixelsProgram;
    supportTimeQuery: boolean;
    private static nextDataId;
    private nextDataId;
    private commandQueueOwnedIds;
    private binaryCache;
    private bufferManager;
    private tensorDisposalQueue;
    private uniformDisposalQueue;
    private disposed;
    private programTimersStack;
    private activeTimers;
    private uploadWaitMs;
    private downloadWaitMs;
    private computePassNumberInEncoder;
    private cpuBackend;
    private querySet;
    constructor(device: GPUDevice, glslang: Glslang, supportTimeQuery?: boolean);
    floatPrecision(): 32;
    flushDisposalQueue(): void;
    /**
     * Dispose the memory if the dataId has 0 refCount. Return true if the memory
     * is released or memory is not managed in this backend, false if memory is
     * not cleared.
     * @param dataId
     * @oaram force Optional, remove the data regardless of refCount
     */
    disposeData(dataId: DataId, force?: boolean): boolean;
    memory(): WebGPUMemoryInfo;
    getBufferManager(): BufferManager;
    acquireBuffer(byteSize: number, usage?: GPUBufferUsageFlags): GPUBuffer;
    maybeReleaseBuffer(dataId: DataId): void;
    /** Return refCount of a `TensorData`. */
    refCount(dataId: DataId): number;
    /** Increase refCount of a `TensorData`. */
    incRef(dataId: DataId): void;
    /** Decrease refCount of a `TensorData`. */
    decRef(dataId: DataId): void;
    write(values: backend_util.BackendValues, shape: number[], dtype: DataType): DataId;
    move(dataId: DataId, values: backend_util.BackendValues, shape: number[], dtype: DataType, refCount: number): void;
    submitQueue(): void;
    getBuffer(dataId: DataId): GPUBuffer;
    ensureCommandEncoderReady(): void;
    private getBufferData;
    private convertAndCacheOnCPU;
    readSync(dataId: object): backend_util.BackendValues;
    read(dataId: object): Promise<backend_util.BackendValues>;
    time(f: () => void): Promise<WebGPUTimingInfo>;
    getAndSavePipeline(key: string, getBinary: () => webgpu_program.WebGPUBinary): webgpu_program.WebGPUBinary;
    makeTensorInfo(shape: number[], dtype: DataType, values?: backend_util.BackendValues | string[]): TensorInfo;
    private tensorToBinding;
    getQueryTime(query: GPUQuerySet): Promise<number>;
    uploadToGPU(dataId: DataId): void;
    runWebGPUProgram(program: webgpu_program.WebGPUProgram, inputs: TensorInfo[], outputDtype: DataType, programUniforms?: Uint32Array | Int32Array | Float32Array): TensorInfo;
    recordFromPixelsCommands(output: GPUBuffer): void;
    getTimeFromQuerySet(querySet: GPUQuerySet): Promise<number>;
    private makeUniforms;
    private getCPUBackend;
    shouldExecuteOnCPU(inputs: TensorInfo[], sizeThreshold?: number): boolean;
    mapActivationToShaderProgram(activation: backend_util.Activation, packed?: boolean): string;
    numDataIds(): number;
    dispose(): void;
}
export {};
